﻿namespace HeuristicLab.Algorithms.PushGP.Expressions {
  using System;
  using System.Collections.Generic;
  using System.Linq;
  using System.Text;
  using HeuristicLab.Algorithms.PushGP.Interpreter;

  public class ExecExpandExpression : StatefullExpression<Expression[]> {
    private const string PREFIX = "( ";
    private const string POSTFIX = ")";
    private const string DELIMITER = " ";

    public readonly int Depth;

    public static readonly ExecExpandExpression Empty = new ExecExpandExpression(new Expression[0]);

    public bool IsEmpty { get; private set; }

    private int[] treeIndex;

    public ExecExpandExpression(Expression[] state)
      : base(state) {
      this.IsEmpty = this.State.Length == 0;
      this.InitId();

      var subExpressions = state.Where(e => e.GetType() == typeof(ExecExpandExpression));

      this.Depth = subExpressions.Any() ? subExpressions.Select(e => e as ExecExpandExpression).Max(e => e.Depth) : 1;
    }

    //private void CheckIfNested(int level, Expression parent)
    //{
    //    for (var i = 0; i < State.Length; i++)
    //    {
    //        if (State[i].GetType() == typeof(ExecExpandExpression))
    //        {
    //            if (parent == State[i])
    //            {
    //                Debugger.Break();
    //            }

    //            (State[i] as ExecExpandExpression).CheckIfNested(level + 1, parent);
    //        }
    //    }
    //}

    public ExecExpandExpression(ExecExpandExpression origin) : base(new Expression[origin.State.Length]) {
      Array.Copy(origin.State, this.State, origin.State.Length);

      this.IsEmpty = origin.IsEmpty;
    }

    public IReadOnlyList<Expression> Expressions { get { return this.State; } }

    /// <summary>
    /// Returns the amount of elements in the whole tree
    /// </summary>
    /// <returns></returns>
    public int TotalCount
    {
      get
      {
        if (this.IsEmpty) {
          return 1;
        }

        this.BuildIndexIfNecessary();

        // + 1 because this is also counted
        return this.treeIndex[0] + 1;
      }
    }

    protected override string InitStringRepresentation() {
      var sb = new StringBuilder();

      sb.Append(PREFIX);
#if DEBUG
      for (var i = this.State.Length - 1; i >= 0; i--)
        sb.Append(this.State[i] + DELIMITER);
#else
      sb.Append(this.TotalCount.ToString() + DELIMITER);
#endif

      sb.Append(POSTFIX);

      return sb.ToString();
    }

    protected sealed override int InitId() {
      var hash = 19 * 31 + this.GetType().GetHashCode();

      for (var i = this.State.Length - 1; i >= 0; i--) hash = hash * 31 + this.State[i].GetHashCode();

      return hash;
    }

    public ExecExpandExpression Copy() {
      // as the empty list is a static stateless expression, no copy required
      return this == Empty || this.IsEmpty
        ? Empty
        : new ExecExpandExpression(this.State);
    }

    public Expression[] CopyExpressions() {
      var copy = new Expression[this.State.Length];
      Array.Copy(this.State, copy, this.State.Length);

      return copy;
    }

    public override int GetHashCode() {
      return base.GetHashCode();
    }

    public override string ToString() {
      return this.StringRepresentation;
    }

    public override bool Equals(object obj) {
      if ((obj == null) || (obj.GetType() != typeof(ExecExpandExpression))) return false;

      return this.GetHashCode() == obj.GetHashCode();
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      interpreter.ExecStack.Push(this.State);
    }

    /// <summary>
    ///     Returns the element with the given index whereby the tree is indexed in depth first manner
    /// </summary>
    /// <param name="index">The index of the required element</param>
    /// <returns>The required element</returns>
    public Expression GetFromTree(int index) {
      return this.GetFromTree(index, (root, parent, child, childIndex, parentIndex) => child);
    }

    /// <returns>The required element</returns>
    public T GetFromTree<T>(int index, Func<ExecExpandExpression, ExecExpandExpression, Expression, int, int, T> resolver) {
      return this.GetFromTree(index, 0, null, resolver);
    }

    private T GetFromTree<T>(int index, int parentIndex, ExecExpandExpression parent, Func<ExecExpandExpression, ExecExpandExpression, Expression, int, int, T> resolver) {
      if (index == 0) return resolver(parent, parent, this, 0, parentIndex);

      this.BuildIndexIfNecessary();

      var min = this.State.Length - 1;
      var max = 0;
      var mid = (min + max) / 2;

      while ((max <= min) && (this.treeIndex[mid] != index)) {
        if (this.treeIndex[mid] > index) max = mid + 1;
        else min = mid - 1;

        mid = (min + max) / 2;
      }

      return this.treeIndex[mid] == index
        ? resolver(parent, this, this.State[mid], mid, parentIndex)
        : (this.State[mid + 1] as ExecExpandExpression).GetFromTree(index - this.treeIndex[mid + 1], mid + 1, this, resolver);
    }

    private void BuildIndexIfNecessary() {
      if (this.treeIndex != null) return;

      this.BuildIndex();
    }

    private void BuildIndex() {
      this.treeIndex = new int[this.State.Length];

      var next = 1;

      for (var i = this.State.Length - 1; i >= 0; i--) {
        var subExpression = this.State[i];

        this.treeIndex[i] = next;

        if (subExpression.GetType() == typeof(ExecExpandExpression)) {
          next += (subExpression as ExecExpandExpression).TotalCount;
        } else {
          next++;
        }
      }
    }
  }

  /// <summary>
  /// If the top item of the BOOLEAN stack is TRUE then this removes the second item on the EXEC stack, leaving the first
  /// item to be executed. If it is false then it removes the first item, leaving the second to be executed. This is similar to
  /// CODE.IF except that it operates on the EXEC stack. This acts as a NOOP unless there are at least two items on the EXEC stack and
  /// one item on the BOOLEAN stack.
  /// </summary>
  public class ExecIfExpression : StatelessExpression {
    protected override string InitStringRepresentation() {
      return "EXEC.IF";
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      // not enough arguments on stack
      if ((interpreter.BooleanStack.Count == 0) || (interpreter.ExecStack.Count < 2)) return;

      var condition = interpreter.BooleanStack.Pop();

      if (condition) interpreter.ExecStack.RemoveAt(interpreter.ExecStack.Count - 2);
      else interpreter.ExecStack.RemoveTop();
    }
  }

  /// <summary>
  ///     Inserts beneath the top item of the EXEC stack a new item of the form
  ///     "( EXEC.Y <TopItem> )".
  /// </summary>
  public class ExecYExpression : StatelessExpression {
    protected override string InitStringRepresentation() {
      return "EXEC.Y";
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      // not enough arguments on stack
      if ((interpreter.ExecStack.Count == 0) || (interpreter.Configuration.MaxPointsInProgram < 2)) return;

      var top = interpreter.ExecStack.Top;
      var result = new ExecExpandExpression(new[] { top, new ExecYExpression() });

      interpreter.ExecStack.Insert(interpreter.ExecStack.Count - 1, result);
    }
  }

  /// <summary>
  ///     Removes the second item on the EXEC stack.
  /// </summary>
  public class ExecKExpression : StatelessExpression {
    protected override string InitStringRepresentation() {
      return "EXEC.K";
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 2) return;

      var top = interpreter.ExecStack.Pop();
      interpreter.ExecStack.SetTop(top);
    }
  }

  /// <summary>
  ///     Pops 3 items from the EXEC stack, which we will call A, B, and C
  ///     (with A being the first one popped). Then pushes a list containing B and C back onto the EXEC stack, followed by
  ///     another instance of C, followed by another instance of A.
  /// </summary>
  public class ExecSExpression : StatelessExpression {
    protected override string InitStringRepresentation() {
      return "EXEC.S";
    }

    public override void Eval(IPushGpInterpreter interpreter) {
      if (interpreter.ExecStack.Count < 3) return;

      var State = interpreter.ExecStack.Pop(2);
      var a = State[1];
      var b = State[0];
      var c = interpreter.ExecStack.Top;

      var newTop = new ExecExpandExpression(new[] { c, b });

      interpreter.ExecStack.SetTop(newTop);
      interpreter.ExecStack.Push(c, a);
    }
  }
}